##################
# deblur testing
##################

module TV

# Our exports
export generate_table

# Dependencies
using FileIO
using ColorTypes: Gray
using Printf
# ColorVectorSpace is only needed to ensure that conversions
# between different ColorTypes are defined.
import ColorVectorSpace
import TestImages
using DelimitedFiles
using ColorSchemes

using AlgTools.Util
using ImageTools.TVRecon
using ImageTools.ImFilter
using ImageTools.ColourTools

# Parameters
const default_params = (
    τ₀ = 5,
    σ₀ = 0.99/5,
    ρ = 0,
    accel = false,
    verbose_iter = 1000,
    image_name = "lighthouse",
    outfile = "tv-results.txt"
)

function rescale(f, δ, δ₀, δ₁, a₀, a₁)
    x, x₀, x₁=f(δ), f(δ₀), f(δ₁)
    return a₀+(x-x₀)/(x₁-x₀)*(a₁-a₀)
end

iteratef = (n, f, x) -> n==1 ? f(x) : f(iteratef(n-1, f, x))

tv_setup = δ -> begin
    α = 0.5*δ # Based on manual experiments
    (δ = δ, α = α, N = ceil(Integer, 100+iteratef(1000, x -> log(1+x), 1/δ)/α))
end

experiments_tv = [ tv_setup(f*10.0^(-p)) for p ∈ 0:8 for f ∈ [1, 0.5]]
experiments_tv_images_only = [ tv_setup(f*10.0^(-p)) for p ∈ [1, 5, 7] for f ∈ 1]

function add_noise(b, δ)
    return b .+ δ.*randn(size(b)...)
end

function quality(b, b_recon)
    return norm₂(b_recon.-b)/norm₂(b)
end

#errorcmap = range(colorant"black", stop=colorant"red", length=256)
#errorimg = d -> mapped_img(abs.(d).^(1/3), errorcmap)
errorcscheme = colorschemes[:coolwarm]
errorimg = d -> get(errorcscheme, abs.(d).^(1/4))

function deblurring_quality(b, blur, b_δ, params, α_δ, Ν_δ, δ)
    x, y, st = recon_pdps(b_δ, blur;
                          xinit = zeros(size(b)...),
                          params = params ⬿ (maxiter = Ν_δ, α = α_δ))
    fn = (t, ext) -> "$(t)_$(δ)_$(Ν_δ).$(ext)"
    save(File(format"PNG", fn("reco", "png")), grayimg(x))
    save(File(format"PNG", fn("error", "png")), errorimg(x.-b))
    return quality(b, x)
end

function generate_table(; images_only=false)
    params = default_params
    b = Float64.(Gray{Float64}.(TestImages.testimage(params.image_name)))
    blur = FilterKernel(gaussian(2, [7 7]))

    results=[]
    
    experiments = images_only ? experiments_tv_images_only : experiments_tv

    for e ∈ experiments
        print("TV deblur $(e)\n")

        b_δ = blur(add_noise(b, e.δ))
        fn = (t, ext) -> "$(t)_$(e.δ).$(ext)"
        save(File(format"PNG", fn("data", "png")), grayimg(b_δ))
        save(File(format"PNG", fn("error", "png")), errorimg(b_δ.-b))

        q_corrupt = quality(b, b_δ)
        if !images_only
            q10 = deblurring_quality(b, blur, b_δ, params, e.α, 10, e.δ)
        end
        q100 = deblurring_quality(b, blur, b_δ, params, e.α, 100, e.δ)
        if !images_only
            q1000 = deblurring_quality(b, blur, b_δ, params, e.α, 1000, e.δ)
            push!(results, (e..., q_corrupt, q, q10, q100, q1000))
        end
        q = deblurring_quality(b, blur, b_δ, params, e.α, e.N, e.δ)
    end

    if !images_only
        print("Writing results to $(params.outfile)\n")
        open(params.outfile, "w") do io
            writedlm(io, ["delta" "alpha" "N" "q_corrupt" "q" "q10" "q100" "q1000"])
            writedlm(io, results)
        end
    end
end

end
